/*
	HTTPGet: Retrieves files via HTTP/1.0 protocol
	by Mikael Kalms (mikael@kalms.org), 2003-04-28

	This program should compile using MSVC++ 6.0.
	It should be easy to make it compile under SAS/C 6.50 as well (most of the
	necessary #ifdefs are already in place).
*/

#ifdef _WIN32

#include <winsock2.h>
#include <windows.h>

#define GETSOCKETERROR WSAGetLastError

#else

#include <string.h>

#include <netdb.h>
#include <proto/socket.h>
#include <sys/socket.h>

#define GETSOCKETERROR Errno

typedef long SOCKET;

#define SOCKET_ERROR -1
#define INVALID_SOCKET -1

#endif


#include <stdio.h>


#define BLKSIZE 1024

#define REPORTBLOCKSIZE (100 * 1024)


char sourceURL[1024];
char sourceProtocol[256];
char sourceHostName[256];
int sourceHostPort;
char sourceFile[1024];

char sourceFileEncoded[1024*3];

char destFile[256];


SOCKET communicationsSocket;

int verbose = 1;

// Find the first occurrence of any of the characters in "pattern" inside "str"

char* strmchr(char* str, char* pattern)
{
	char* s = str;

	if (!pattern[0])
		return str;

	for (; *s; s++)
	{
		char* p = pattern;
		for (; *p; p++)
			if (*s == *p)
				return s;
	}

	return 0;
}

// Establish connection to web server

int connectToWebServer(void)
{
	struct sockaddr_in addr;
	int ret;
	unsigned long ip;
	unsigned long ip2;
	struct hostent* hostEnts;

	// First, attempt to look up server name in DNS database

	if (!(hostEnts = gethostbyname(sourceHostName)))
	{
		// If not found, attempt to convert the name into a raw IP address
		
		if (verbose >= 2)
			printf("DNS doesn't know about address \"%s\", try it as an IP address instead...\n", sourceHostName);

		if ((ip = inet_addr(sourceHostName)) == INADDR_NONE)
		{
			if (verbose >= 2)
				printf("\"%s\" is not a valid IP address either\n", sourceHostName);

			fprintf(stderr, "Error: %s: No such host\n", sourceHostName);
			return 0;
		}

		ip2 = ntohl(ip);
	}
	else
	{
		// Server has been found in DNS

		ip = *((unsigned long*) hostEnts->h_addr);
		ip2 = ntohl(ip);

		if (verbose >= 2)
			printf("Found \"%s\" in DNS; the IP is %d.%d.%d.%d\n", sourceHostName, (ip2 >> 24) & 0xff, (ip2 >> 16) & 0xff, (ip2 >> 8) & 0xff, ip2 & 0xff);
	}

	memset(&addr, 0, sizeof(addr));

#ifndef _WIN32
	addr.sin_len = 0;
#endif
	addr.sin_family			= AF_INET;
	addr.sin_port			= htons(sourceHostPort);
	addr.sin_addr.s_addr	= ip;

	// Create socket

	if (verbose >= 2)
		printf("Attempting to create socket...\n");

	if ((communicationsSocket = socket( AF_INET, SOCK_STREAM, 0 )) == INVALID_SOCKET)
	{
		fprintf(stderr, "socket: %d\n", GETSOCKETERROR());
		return 0;
	}

	if (verbose >= 2)
		printf("socket OK\n");

	// Attempt to connect socket to web server

 	if (verbose >= 2)
		printf("Attempting to establish connection with web server at %d.%d.%d.%d:%d...\n", (ip2 >> 24) & 0xff, (ip2 >> 16) & 0xff, (ip2 >> 8) & 0xff, ip2 & 0xff, sourceHostPort);

	if ((ret = connect(communicationsSocket, (struct sockaddr*)&addr, sizeof(struct sockaddr))) == SOCKET_ERROR)
	{
		fprintf(stderr, "connect: %d\n", GETSOCKETERROR());
		return 0;
	}

	if (verbose >= 2)
		printf("connect OK\n");

	return 1;
}

// Shut down connection to web server

void disconnectFromWebServer(void)
{
#ifdef _WIN32
	shutdown(communicationsSocket, SD_SEND);

	closesocket(communicationsSocket);
#else
	CloseSocket(communicationsSocket);
#endif
}

// Attempt to send the specified block to the web server

int sendBlk(void* buf, int blkSize)
{
	int ret;
	if ((ret = send(communicationsSocket, buf, blkSize, 0)) == SOCKET_ERROR)
	{
		fprintf(stderr, "Error: send: %d\n", GETSOCKETERROR());
		return 0;
	}

	if (!ret)
	{
		fprintf(stderr, "Error: No response from the other side\n");
		return 0;
	}

	return 1;
}

// Wait for a block of data to be received from the web server

int recvBlk(void* buf, int blkSize)
{
	int ret;
	if ((ret = recv(communicationsSocket, buf, blkSize, 0)) == SOCKET_ERROR)
	{
		fprintf(stderr, "Error: recv: %d\n", GETSOCKETERROR());
		return -1;
	}

	return ret;
}

// Encode filename according to RFC1738

void encodeFileName(char* dest, char* src)
{
	char hex2asc[16] = "0123456789ABCDEF";

	while (*src)
	{
		if (!strchr("$-_.+!*'(),/:?", *src)
			&& (*src <= 0x2e
				|| (*src >= 0x3a && *src <= 0x40)
				|| (*src >= 0x5b && *src <= 0x60)
				|| *src >= 0x7b))
		{
			*dest++ = '%';
			*dest++ = hex2asc[(*src >> 4) & 0x0f];
			*dest++ = hex2asc[*src & 0x0f];
			src++;
		}
		else
			*dest++ = *src++;
	}

	*dest++ = 0;
}

int main(int argc, char** argv)
{
#ifdef _WIN32
	WSADATA data;
#endif

	int ret;
	int arg;
	int nonEscapedArg = 0;
	int help = 0;
	char getRequest[1024*3];
	int totalRetrievedBytes = 0;


	memset(sourceURL, 0, sizeof(sourceURL));
	memset(sourceProtocol, 0, sizeof(sourceProtocol));
	memset(sourceHostName, 0, sizeof(sourceHostName));
	sourceHostPort = 80;
	memset(sourceFile, 0, sizeof(sourceFile));
	memset(destFile, 0, sizeof(destFile));

	// Parse commandline arguments

	for (arg = 1; arg < argc; arg++)
	{
		if (!stricmp(argv[arg], "-q") || !stricmp(argv[arg], "--quiet"))
			verbose = 0;
		else if (!stricmp(argv[arg], "-v") || !stricmp(argv[arg], "--verbose"))
			verbose = 2;
		else if (!stricmp(argv[arg], "-h") || !stricmp(argv[arg], "--help"))
			help = 1;
		else
		{
			if (nonEscapedArg == 0)
			{
				char* protocol = argv[arg];
				char* host;
				char* port;
				char* file;

				// Find protocol specification

				host = strstr(protocol, "://");
				if (!host)
				{
					host = protocol;
					protocol = 0;
				}
				else
					host += 3;

				if (protocol)
					memcpy(sourceProtocol, protocol, host - protocol - 3);
				else
					strcpy(sourceProtocol, "http");

				// Find file specification
				if (!(file = strchr(host, '/')))
					strcpy(sourceFile, "/");
				else
					strcpy(sourceFile, file);

				// Find port specification, if there is any
				port = strmchr(host, ":/");
				if (port && (port != file))
					sourceHostPort = atoi(port + 1);

				// Find host specification
				if (port)
					memcpy(sourceHostName, host, port - host);
				else if (file)
					memcpy(sourceHostName, host, file - host);
				else
					strcpy(sourceHostName, host);

			}
			else if (nonEscapedArg == 1)
			{
				strcpy(destFile, argv[arg]);
			}
			nonEscapedArg++;
		}
	}

	if (!sourceHostName[0] || !sourceFile[0] || !destFile[0])
		help = 1;

	// If necessary, display help information

	if (help)
	{
		printf("\nHTTPGet: Retrieves files via HTTP protocol\n");
		printf("By Mikael Kalms (mikael@kalms.org) 2003-04-28\n");
		printf("\n");
		printf("Usage:   HTTPGet [options] <source URL> <destination filename>\n");
		printf("\n");
		printf("Options: -h,--help       Display this help text\n");
		printf("         -q,--quiet      Only output error messages\n");
		printf("         -v,--verbose    Output debug info\n");
		printf("\n");
		printf("Example: HTTPGet -q www.altavista.com/index.html MyPage.html\n");
		return 0;
	}

#ifdef _WIN32
	// Startup WinSock on win32

	if ((ret = WSAStartup( MAKEWORD( 2, 2 ), &data )) == SOCKET_ERROR)
	{
		fprintf(stderr, "Error: WSAStartup: %d\n", WSAGetLastError());
		return 1;
	}

	if (verbose >= 2)
		printf("WSAStartup OK\n");
#endif

	// Connect to web server

	if (verbose >= 1)
		printf("Attempting to connect to host \"%s\" port %d...\n", sourceHostName, sourceHostPort);

	if (!connectToWebServer())
	{
		fprintf(stderr, "Error: unable to connect to web server\n");
		return 0;
	}

	if (verbose >= 1)
		printf("Connection to web server established\n");

	// Post HTTP GET request

	encodeFileName(sourceFileEncoded, sourceFile);
	sprintf(getRequest, "GET %s://%s%s HTTP/1.0\015\012\015\012", sourceProtocol, sourceHostName, sourceFileEncoded);

	if (verbose >= 2)
		printf("Sending HTTP GET request for page \"%s://%s%s\" to web server...\n", sourceProtocol, sourceHostName, sourceFileEncoded);

	if (!sendBlk(getRequest, strlen(getRequest)))
	{
		fprintf(stderr, "Error: Unable to send HTTP GET request to the web server\n");
		disconnectFromWebServer();
		return 0;
	}

	if (verbose >= 2)
		printf("HTTP GET request has successfully been sent\n");

	// Receive response

	if (verbose >= 2)
		printf("Awaiting response from web server...\n");

	while (1)
	{
		static int firstPacket = 1;
		static char buf[BLKSIZE];
		char* bufStart;
		
		int bytesRead = recvBlk(buf, sizeof(buf));

		if (bytesRead == -1)
		{
			disconnectFromWebServer();
			return 0;
		}
		else if (bytesRead > 0)
		{
			FILE* f;
			if (firstPacket)
			{
				static char seg1[256];
				static char seg2[256];

				if (verbose >= 1)
					printf("Retrieved first packet from web server\n");

				// Here, we assume that the entire header will fit within one response packet

				if (sscanf(buf, "%s %s", seg1, seg2) != 2
					|| memcmp(buf, "HTTP/", 5)
					|| !(bufStart = strstr(buf, "\015\012\015\012")))
				{
					fprintf(stderr, "Error: Web server responded with an invalid page (maybe it only talks HTTP/0.9 protocol?)\n");
					disconnectFromWebServer();
					return 0;
				}

				if (stricmp(seg2, "200"))
				{
					int cause = atoi(seg2);
					char* causeStr = "";
					switch (cause)
					{
					case 400:
						causeStr = "(Web server does not understand the format of the HTTP GET request)";
						break;
					case 401:
						causeStr = "(You must authenticate yourself (username/password) to access this page)";
						break;
					case 403:
						causeStr = "(Forbidden -- web server refuses to yield any more data)";
						break;
					case 404:
						causeStr = "(File not found)";
						break;
					case 500:
						causeStr = "(Internal web server error)";
						break;
					case 503:
						causeStr = "(Web server is temporarily unavailable)";
						break;

					};
					fprintf(stderr, "Error: Web server says, error %s %s\n", seg2, causeStr);
					disconnectFromWebServer();
					return 0;
				}

				bufStart += 4;

				f = fopen(destFile, "wb");
			}
			else
				bufStart = buf;
				f = fopen(destFile, "ab");

			bytesRead -= (bufStart - buf);

			if (!f)
			{
				fprintf(stderr, "Error: Unable to open output file \"%s\"\n", destFile);
				disconnectFromWebServer();
				return 0;
			}

			fwrite(bufStart, bytesRead, 1, f);
			fclose(f);

			if (verbose >= 1
				&& totalRetrievedBytes / REPORTBLOCKSIZE != (totalRetrievedBytes + bytesRead) / REPORTBLOCKSIZE)
				printf("Retreived %d bytes from web server\n", totalRetrievedBytes + bytesRead);

			totalRetrievedBytes += bytesRead;

			firstPacket = 0;
		}
		else
			break;
	}

	disconnectFromWebServer();

	if (verbose >= 1)
		printf("Retrieved %d bytes into file \"%s\"\n", totalRetrievedBytes, destFile);

	if (verbose >= 1)
		printf("Done\n");
	
	return 0;
}
